//
// BrackRecVirth.go - рекурсивный распознаватель вложенности круглых скобок.
// Построен на основе диаграмм Вирта, эквивалентных следующей РБНФ
//
// $S = {/ "(" [ S ] ")" /}.
//

package main

import (
    "fmt"
    "strings"
)

var str string      // Строка с входной цепочкой, имитирующая входную ленту
var i int           // Текущее положение входной головки
var brackets []rune // Срез, содержащий прочитанную строку со скобками

var erFlag int          // Флаг, фиксирующий наличие ошибок в середине правила

// Функция, реализующая чтение символов в буфер из входного потока.
// Используется для ввода с клавиатуры распознаваемой строки.
// Ввод осуществляется до нажатия на клавишу Enter.
// Символ '\n' яляется концевым маркером входной строки.
func GetOneLine() {
    fmt.Scanln(&str)
    str += "\n"; // Добавляется концевой маркер
}

// Функция, реализующая распознавание нетерминала S.
func S() bool {
//_0: // Начало диаграммы
    if(brackets[i] == '(') {
        i++;
        goto _1
    }
    return false    // Первый символ цепочки некорректен
                    // что это ошибка, лучше определить снаружи
  _1: // Точка 1 диаграммы
        if(brackets[i] == ')') {
            i++
            goto _3
        }
        if(S()) {
            goto _2
        }
        erFlag++
        fmt.Print("Position ", i, ", ",
                  "Error 1: I want closed bracket or next opened bracket!\n",
            )
        return false
  _2: // Точка 2 диаграммы
        if(brackets[i] == ')') {
            i++
            goto _3
        }
        erFlag++
        fmt.Print("Position ", i, ", ",
                  "Error 2: I want closed bracket!\n",
            )
        return false
  _3: // Точка 3 диаграммы
        if(brackets[i] == '(') {
            i++
            goto _1
        }
        goto _end;
  _end: // Точка end диаграммы
    return true;
}

// Функция, реализующая распознавание нетерминала Z.
func Z() bool {
//_0: // Начало диаграммы
    if(S()) { goto _1 }
    return false    // Первый символ цепочки некорректен
                    // что это ошибка, лучше определить снаружи
_1: // Точка 1 диаграммы
    // За последней скобкой должен быть "конец строки"
    if(brackets[i] == '\n') {
        goto _end   // Все прошло нормально
    }
    erFlag++
    fmt.Print("Position ", i, ", ",
              "Error 3: I want end of line!\n",
        )
    return false
_end:
    return true
}

// Функция запускающая разбор и определяющая корректность его завершения,
// если первый символ не принадлежит цепочки
func Parser() bool {
    // Начальная инициализация.
    erFlag = 0
    i = 0;

    // Процесс пошел!
    if(Z()) {
        return true     // Все прошло нормально
    } else {
        if(erFlag != 0) {
            fmt.Print("Position ", i, ", ",
                      "Error 4: Internal Error!\n",
                )
        } else {
            fmt.Print("Position ", i, ", ",
                      "Error 5: Incorrect first symbol of S!\n",
                )
        }
        return false; // Есть ошибки
    }
}

//=============================================================================
// Главная функция используется для тестирования до тех пор, 
// пока не будет прочитана пустая строка
func main() {
    // Цикл распознавания различных входных цепочек
    for {
        strCursor := ""
        str = "";
        
        // Чтение очередной входной цепочки в буфер
        fmt.Print("Input bracket's expression: ")

        // Формируем очередную строку скобок для распознавания.
        GetOneLine()
        brackets = []rune(str)

        // Здесь начинается разбор принятой строки.
        if Parser() {
            fmt.Print("+++++ OK! +++++\n")
        } else {
            fmt.Print("----- Fatal error (look upper error message)! -----\n")
        }

        // Вывод разобранной строки и значения позиции входой головки.
        fmt.Print("Line: ", str)
        strCursor = " Pos: " + strings.Repeat("-", i)
        strCursor +="^"
        fmt.Print(strCursor, "  i = ", i, "\n\n")

        if str == "\n" {
            break
        }
    }
    fmt.Println("Goodbye!")
}
